//		Copyright © 1999	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Screen saver class.

#include <cmath>
#include <cstdlib>
#include <cstring>

#include <GraphicsDefs.h>
#include <image.h>
#include <Message.h>
#include <OS.h>
#include <Point.h>
#include <Rect.h>
#include <ScreenSaver.h>
#include <TextView.h>
#include <View.h>

#include "SuckSaver.h"
#include "RandomColor.h"

// Background color.
const rgb_color kBlack = { 0, 0, 0, 0 };

// Parameters governing the behavior.
const float kFramesPerSecond = 75.0;
const float kInitialRevsPerSecond = 0.1;
const float kEventHorizon = 5.0;
const float kTimeToHorizon = 4.0;
const float kRadiusSpread = 0.1;
const float kVelocitySpread = 0.05;

// Random-float generator.
inline float
randomFloat(const float min = 0.0, const float max = 1.0)
{
	return min + ((float)rand() / (RAND_MAX+1)) * (max - min);
}

// Screen saver interface.
BScreenSaver *
instantiate_screen_saver(BMessage *pMsg, image_id id)
{
	return new SuckSaver(pMsg, id);
}

SuckSaver::SuckSaver(BMessage *pMsg, image_id id)
:	inherited(pMsg, id),
	mSucker(B_ORIGIN),
	mInitialRadius(0),
	mInitialVelocity(0),
	mGravity(0),
	mNumParticles(0)
{
	srand((long)system_time());
	for (int i = 0; i < kMaxParticles; ++i) {
		mpParticles[i] = 0;
	}
}

SuckSaver::~SuckSaver(void)
{
	for (int i = 0; i < kMaxParticles; ++i) {
		delete mpParticles[i]; mpParticles[i] = 0;
	}
}

void
SuckSaver::Draw(BView *pView, int32 frame)
{
	pView->SetLowColor(kBlack);
	if (frame == 0) {
		pView->FillRect(pView->Bounds(), B_SOLID_LOW);
	}
	
	for (int i = 0; i < mNumParticles; ++i) {
		if (!mpParticles[i] && randomFloat() > 0.98) {
			// Create a new particle.
			const float angle = 2.0 * M_PI * randomFloat();
			const float radius =
				mInitialRadius * randomFloat(1.0-kRadiusSpread, 1.0+kRadiusSpread);
			const float velocity =
				mInitialVelocity * randomFloat(1.0-kVelocitySpread, 1.0+kVelocitySpread);
			mpParticles[i] = new Particle(angle, radius, velocity);
		}
		
		if (mpParticles[i]) {
			Particle &p = *mpParticles[i];
			
			// Erase the particle.
			const BPoint pos = Pos(p);
			const BPoint offset(2, 2);
			pView->FillRect(BRect(pos - offset, pos + offset), B_SOLID_LOW);
			
			// Move the particle.
			Transform(p);

			if (mpParticles[i]->mRadius < kEventHorizon) {
				// Particle has expired.
				delete mpParticles[i];
				mpParticles[i] = 0;
				
			} else {
				// Redraw it.
				pView->SetHighColor(mpParticles[i]->mColor);
				const BPoint pos = Pos(p);
				const BPoint offset(2, 2);
				pView->FillRect(BRect(pos - offset, pos + offset), B_SOLID_HIGH);
			}
		}
	}
}

void
SuckSaver::StartConfig(BView *pConfigView)
{
	const BRect frame = pConfigView->Bounds().InsetBySelf(10.0, 10.0);
	BTextView *pCaption =
		new BTextView(frame, "SuckSaver", frame.OffsetToCopy(B_ORIGIN),
							B_FOLLOW_ALL, B_WILL_DRAW);
	pCaption->SetText("SuckSaver\n"
							"\n"
							"\t© 1999  Jens Kilian <jjk@acm.org>");
	pConfigView->AddChild(pCaption);
	pCaption->SetViewColor(pConfigView->ViewColor());
	pCaption->SetStylable(true);
	pCaption->SetFontAndColor(be_plain_font);
	const char *pText = pCaption->Text();
	pCaption->SetFontAndColor(0, strchr(pText, '\n') - pText, be_bold_font);
	pCaption->MakeEditable(FALSE);
	pCaption->MakeSelectable(FALSE);
}

status_t
SuckSaver::StartSaver(BView *pView, bool preview)
{
	// Calculate refresh rate.
	if (preview) {
		SetTickSize((bigtime_t)(1.0e5 / kFramesPerSecond));
		mNumParticles = kMaxParticles / 10;
	} else {
		SetTickSize((bigtime_t)(1.0e6 / kFramesPerSecond));
		mNumParticles = kMaxParticles;
	}

	// Calculate parameters of the model from the view size.
	const BRect bounds = pView->Bounds();
	mSucker = BPoint((bounds.left + bounds.right) * 0.5,
						  (bounds.top + bounds.bottom) * 0.5);
	
	const float width = bounds.Width();
	const float height = bounds.Height();
	mInitialRadius = hypot(width, height) * 0.5;
	if (mInitialRadius < kEventHorizon * 5.0) {
		// View is too small.
		return B_ERROR;
	}
	
	mInitialVelocity = 2.0 * M_PI * mInitialRadius
						  * kInitialRevsPerSecond
						  / kFramesPerSecond;

	mGravity = (mInitialRadius - kEventHorizon)
				/ (preview ? kTimeToHorizon * 5.0 : kTimeToHorizon)
				/ kFramesPerSecond;

	return B_OK;
}

void
SuckSaver::Transform(SuckSaver::Particle &p) const
{
	p.mAngle += p.mVelocity / p.mRadius;
	p.mRadius -= mGravity;
}

BPoint
SuckSaver::Pos(const SuckSaver::Particle &p) const
{
	return mSucker + BPoint(p.mRadius*cos(p.mAngle), p.mRadius*sin(p.mAngle));
}
